from pytest import Session, Item, Parser
from _pytest.fixtures import FixtureDef, FixtureValue, SubRequest
from datetime import datetime
from loguru import logger
from handshake.services.DBService.lifecycle import log_less
from handshake.reporters.pytest_reporter import PyTestHandshakeReporter, PointToAtPhase

reporter = PyTestHandshakeReporter()


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "test_metadata(title: str, description: str): Attach metadata to handshake like title and description to tests.",
    )


def pytest_addoption(parser: Parser):
    parser.addini(
        "save_results_in",
        "saves the test results [db file] in the give path [note this is not where the html export would be generated]",
        type="string",
    )
    parser.addini(
        "save_handshake_config_dir",
        "saves the handshake configuration in this path, once generated. you can configure the reporter",
        type="string",
    )
    parser.addini(
        "disable_handshakes",
        "disable handshake reporter, set it as either 1/true or false",
        type="bool",
    )
    parser.addini(
        "handshake_port",
        "port number that handshake should use to collect test results",
        type="string",
    )


def pytest_sessionstart(session: Session):
    if reporter.parse_config(session):
        return
    is_quiet = session.config.getoption("q", session.config.getoption("quiet", False))
    if is_quiet:
        log_less()

    reporter.start_collection(
        session.config.inicfg.get("projectName") or session.path.name, is_quiet
    )
    reporter.create_session(datetime.now())
    reporter.put_test_config(session.config.inicfg)


def pytest_itemcollected(item: Item):
    reporter.create_test_entity(item)


def pytest_runtest_logstart(nodeid, location):
    reporter.update_test_entity_details(None, nodeid)


def pytest_runtest_setup(item: Item):
    reporter.create_test_entity(item, helper_entity=PointToAtPhase.SETUP)


def pytest_runtest_call(item: Item):
    reporter.func_args[item.nodeid] = item.funcargs


def pytest_runtest_logreport(report):
    reporter.update_test_entity_details(report)


def pytest_runtest_teardown(item: Item, nextitem: Item):
    reporter.create_test_entity(item, helper_entity=PointToAtPhase.TEARDOWN)


#
def pytest_assertion_pass(item, lineno, orig, expl):
    try:
        reporter.add_test_assertion(item.nodeid, orig, expl, True)
    except Exception:
        logger.exception("Unable to add test assertion")


def pytest_fixture_post_finalizer(
    fixturedef: FixtureDef[FixtureValue], request: SubRequest
):
    reporter.note_fixture(fixturedef, request)


def pytest_sessionfinish(session: Session, exitstatus: int):
    force_call = reporter.update_test_status(session, exitstatus)
    if not force_call:
        reporter.mark_suites_for_processing()
        reporter.update_session(session)
    reporter.close_resources()



from asyncio import gather, to_thread
from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    db_path,
    close_connection,
    attachment_folder,
)
from shutil import rmtree
from pathlib import Path
from typing import Optional
from loguru import logger
from tortoise.expressions import Q, Subquery
from ansitohtml.parser import Parser
from handshake.services.DBService.models.result_base import RunBase
from tortoise import connections, BaseDBAsyncClient
from handshake.services.DBService.models.config_base import (
    ConfigBase,
    ConfigKeys,
)
from handshake.services.SchedularService.register import register_bulk_excel_export
from handshake.services.DBService.models.dynamic_base import TaskBase, JobType
from handshake.services.SchedularService.flag_tasks import pruneTasks
from handshake.services.SchedularService.handlePending import patch_jobs
from handshake.Exporters.json_exporter import JsonExporter
from handshake.Exporters.excel_exporter import excel_export
from handshake.Exporters.html_reporter import HTMLExporter


class Scheduler:
    RUNS_PAGE_VERSION = 1

    def __init__(
        self,
        root_dir: str,
        out_dir: Optional[str] = None,
        manual_reset: Optional[bool] = False,
        inside_test_results: Optional[bool] = False,
        dev: Optional[bool] = False,
        export_mode: str = "json",
        include_excel_export: Optional[bool] = False,
    ):
        self.db_path = db_path(root_dir)
        if not export_mode and export_mode != "json" and export_mode != "html":
            logger.warning(
                "export mode was not defined, please mention it either has html or json."
                " For example, handshake export TestResults -e html",
                export_mode,
            )

        self.exporter = (
            JsonExporter(
                self.db_path,
                (
                    Path(out_dir)
                    if out_dir and not inside_test_results
                    else self.db_path.parent
                ),
                dev or inside_test_results,
            )
            if export_mode == "json"
            else HTMLExporter(
                self.db_path,
                Path(out_dir),
                dev or inside_test_results,
            )
        )
        self.excel_export = include_excel_export
        self.skip_export = out_dir is None and not inside_test_results
        if self.skip_export and self.exporter:
            logger.warning(
                "Calculating Results but please note Export would be skipped,"
                " if you need the results to be exported then please pass export mode and output directory,"
                " for example: handshake export TestResults -e json -o TestReports"
            )
        self.converter = Parser()
        self.db_path = db_path(root_dir)
        self.reset = manual_reset
        self.connection: Optional[BaseDBAsyncClient] = None

    async def rotate_test_run(self, projectName: str, to_delete: int):
        project_runs = RunBase.filter(projectName=projectName).all()
        will_be_deleted = await project_runs.count() - to_delete

        if will_be_deleted <= 0:
            logger.debug(
                "Test Runs for the project: {} has not yet reached the limit,"
                " hence skipping the deletion Task for this project",
                projectName,
            )
            return 0

        to_delete_runs = await (
            project_runs.order_by("started")
            .limit(will_be_deleted)
            .values_list("testID", flat=True)
        )
        await gather(
            RunBase.filter(Q(testID__in=to_delete_runs)).delete(),
            *(
                to_thread(
                    rmtree,
                    attachment_folder(self.db_path, test_id),
                    ignore_errors=False,
                    onerror=None,
                )
                for test_id in to_delete_runs
                if attachment_folder(self.db_path, test_id).exists()
            ),
        )

        logger.info(
            "We have deleted {} test runs under the project name: {}. due to the max. threshold of {} test runs",
            will_be_deleted,
            projectName,
            to_delete,
        )
        return will_be_deleted

    async def rotate_test_runs(self):
        runs_per_project = await ConfigBase.filter(
            key=ConfigKeys.maxRunsPerProject
        ).first()
        requested = max(int(runs_per_project.value), 2)

        recently_deleted = sum(
            await gather(
                *(
                    self.rotate_test_run(projectName, requested)
                    for projectName in await RunBase.all()
                    .distinct()
                    .values_list("projectName", flat=True)
                )
            )
        )

        record = await ConfigBase.filter(key=ConfigKeys.recentlyDeleted).first()
        if record:
            await record.update_from_dict(dict(value=str(recently_deleted)))
            await record.save()
        else:
            await ConfigBase.create(
                key=ConfigKeys.recentlyDeleted, value=str(recently_deleted)
            )

        logger.debug("Delete job is completed.")

    async def init_jobs(self):
        await pruneTasks()

        prev_picked_tasks = await TaskBase.filter(processed=False, picked=True).all()
        reset_from_config = await ConfigBase.filter(
            key=ConfigKeys.reset_test_run
        ).first()

        self.reset = self.reset or reset_from_config.value

        if self.reset:
            logger.debug(
                "we would be processing the test runs again, we can expect updated results."
            )

        to_reset = (
            (Q(type=JobType.MODIFY_TEST_RUN) | Q(type=JobType.EXPORT_EXCEL))
            if self.excel_export and excel_export
            else Q(type=JobType.MODIFY_TEST_RUN)
        )

        to_modify_test_runs = (
            await TaskBase.filter(to_reset & Q(processed=True)).all()
            if self.reset
            else []
        )
        to_pick = prev_picked_tasks + to_modify_test_runs

        async def pick_old_tasks():
            for task in to_pick:
                logger.debug(
                    "This unfinished task: {}[{}] would run in this iteration",
                    task.type,
                    task.ticketID,
                )
                task.picked = False
                task.processed = False

            if to_pick:
                await TaskBase.bulk_update(to_pick, ("picked", "processed"), 100)
                logger.debug("Done!, Marked {} old tasks for processing", len(to_pick))

        async def reset_completed_runs():
            runs = []
            for ticket in to_modify_test_runs:
                test: RunBase = await ticket.test
                logger.debug("Reset for test run: {}", test.testID)
                test.standing = "PENDING"
                runs.append(test)

            if runs:
                logger.debug("Reset done for {} test runs!", len(runs))
                await RunBase.bulk_update(runs, ("standing",), 100)

            if reset_from_config:
                reset_from_config.value = ""
                await reset_from_config.save()

        async def add_export_jobs():
            if not self.excel_export:
                return

            test_ids_to_pick = await RunBase.filter(
                ~Q(
                    testID__in=Subquery(
                        TaskBase.filter(type=JobType.EXPORT_EXCEL).values_list(
                            "test_id", flat=True
                        )
                    ),
                )
            ).values_list("testID", flat=True)

            if not excel_export:
                return logger.warning(
                    "Excel Export due to non-availability of the library, please proceed with the pip install "
                    "handshakes[excel-export] to download the openpyxl which is required for the excel export",
                )

            test_ids_to_pick and await register_bulk_excel_export(test_ids_to_pick)

        await gather(pick_old_tasks(), reset_completed_runs(), add_export_jobs())
        logger.debug("Pre-Patch Jobs have been initiated")

    async def start(self, config_path: Optional[str] = None):
        await init_tortoise_orm(self.db_path, True, config_path=config_path)
        self.connection = connections.get("default")
        await self.rotate_test_runs()
        await self.init_jobs()

        await patch_jobs(self.excel_export, self.db_path)

        if not self.skip_export:
            await self.exporter.start_exporting()

        await close_connection()



from handshake.services.Endpoints.internalEndpoints import one_liners
from handshake.services.DBService.shared import APP_NAME
from handshake.services.Endpoints.blueprints.updateEndpoints import update_service
from handshake.services.Endpoints.blueprints.createService import create_service
from handshake.services.Endpoints.blueprints.writeServices import writeServices
from handshake.services.Endpoints.errorHandling import handle_validation_error
from handshake.services.Endpoints.blueprints.utils import (
    extractPayload,
    attachInfo,
    attachError,
    attachWarn,
)
from sanic import Sanic, Request
from sanic.response import JSONResponse
from pydantic import ValidationError
from sanic.blueprints import Blueprint
from dotenv import load_dotenv

load_dotenv()

listeners = Blueprint.group(
    create_service, update_service, writeServices, name_prefix="listeners"
)


service_provider = Sanic(APP_NAME)


service_provider.config.TOUCHUP = False
service_provider.blueprint(one_liners)
service_provider.blueprint(listeners)

service_provider.error_handler.add(ValidationError, handle_validation_error)


@service_provider.on_response
async def handle_response(request: Request, response: JSONResponse):
    if 200 <= response.status < 300:
        if hasattr(service_provider.shared_ctx, "DEV"):
            await attachInfo(extractPayload(request, response), request.url)
        return response

    payload = extractPayload(request, response)
    if response.status == 400:
        await attachError(payload, request.url)
    else:
        # let's not raise an error whenever someone requests something from this server
        await attachWarn(payload, request.url)

    return JSONResponse(body=payload, status=response.status)



import datetime
from sqlite3 import connect
from handshake.services.DBService import DB_VERSION, OLDEST_VERSION
from handshake.services.DBService.models.enums import ConfigKeys
from handshake.services.DBService.models.config_base import (
    MigrationStatus,
    MigrationTrigger,
)
from sqlite3.dbapi2 import Connection
from typing import Optional, Tuple, Union, Literal
from pathlib import Path
from loguru import logger


def log_migration(
    connection: Connection,
    from_version: int,
    to_version: int,
    trigger: MigrationTrigger,
    passed: bool,
    error: Optional[Exception] = None,
):
    connection.execute(
        "INSERT INTO migrationbase(fromVersion, toVersion, modified, trigger, status, error) VALUES(?,?,?,?,?,?)",
        (
            from_version,
            to_version,
            datetime.datetime.now(tz=datetime.UTC).isoformat(),
            trigger,
            MigrationStatus.PASSED if passed else MigrationStatus.FAILED,
            repr(error) if error else "",
        ),
    )


def check_version(
    connection: Optional[Connection] = None, path: Optional[Path] = None, is_auto=False
) -> Tuple[bool, bool, Union[Literal[False], int]]:
    query = f"select value from configbase where key = '{ConfigKeys.version}'"

    close_connection = not connection
    connection = connection if connection else connect(path)

    try:
        result = connection.execute(query).fetchone()
    finally:
        close_connection and connection.close()

    version_stored = False
    migration_required = False

    if not result:
        logger.warning(
            f"Could not find the version inside the configbase which is not expected!,"
            f" Please raise this as an issue."
        )
    else:
        version_stored = result if not result else int(result[0])
        migration_required = version_stored != DB_VERSION

        logger.log(
            "DEBUG" if not migration_required else "WARNING",
            (
                "Currently at: v{}."
                if not migration_required
                else "Found version: v{}. but required is v{}."
                + (
                    ""
                    if is_auto or version_stored > DB_VERSION
                    else ' Please execute: \n"handshake migrate [COLLECTION_PATH]"'
                )
            ),
            result[0],
            DB_VERSION,
        )
    return (
        migration_required,  # migration is required or now
        version_stored < DB_VERSION,  # bump version or not
        version_stored,  # what was the version stored?
    )


def migration(path: Path, trigger=MigrationTrigger.AUTOMATIC, do_once=False) -> bool:
    if not path.exists():
        logger.debug("Migration check is not required, as the db does not exist.")
        return False
    connection = connect(path)
    stored_version = False
    try:
        is_required, bump_required, stored_version = check_version(
            connection, is_auto=True
        )

        if not is_required:
            logger.debug("Already migrated to required version.")
            return False
        if not stored_version:
            return False
        if not bump_required:
            logger.error(
                "Found a version v{}. but we can migrate till v{}. "
                "Please check the python build version and accordingly "
                'execute: "handshake step-back [COLLECTION_PATH]"',
                stored_version,
                DB_VERSION,
            )
            return False

        for version_to_bump in range(stored_version, DB_VERSION):
            script = Path(__file__).parent / "scripts" / f"bump-v{version_to_bump}.sql"
            logger.info(
                "Executing {} to migrate from v{} to v{}",
                script.name,
                version_to_bump,
                version_to_bump + 1,
            )
            connection.executescript(script.read_text())
            if do_once:
                # this is only for testing purposes
                break
        to_version = (stored_version + 1) if do_once else DB_VERSION

        log_migration(
            connection,
            stored_version,
            to_version,
            trigger,
            True,
        )
        connection.commit()
        logger.info("Migrated to latest version v{}!", to_version)
        return True
    except Exception as error:
        logger.error("Failed to execute migration script, due to {}", error)
        connection.rollback()
        logger.warning(
            "Changes made are rolled back you are now in v{}. Please raise an issue regarding this.",
            stored_version,
        )
        stored_version and log_migration(
            connection,
            stored_version,
            (stored_version + 1) if do_once else DB_VERSION,
            trigger,
            False,
            error,
        )
        connection.commit()
        return False
    finally:
        connection.close()


def revert_step_back(
    to_revert: int,
    db_path: Path,
):
    # we do not recommend calling this function unless it is necessary

    logger.warning(
        "Please note we would be reverting from version: v{} to v{}",
        to_revert,
        to_revert - 1,
    )
    script = Path(__file__).parent / "scripts" / f"revert-v{to_revert}.sql"
    if not script.exists():
        if to_revert <= OLDEST_VERSION:
            logger.error(
                "we didn't find a revert script: {}. you are in the older version of handshakes",
                script.name,
            )
        else:
            logger.error(
                "we didn't find a revert script: {}. you are already in the older version of db",
                script.name,
            )
        logger.error("NOTE: you are now in the version: v{}", to_revert)
        return False

    connection = connect(db_path)
    try:
        connection.executescript(script.read_text())
        log_migration(connection, to_revert, to_revert - 1, MigrationTrigger.CLI, True)
        connection.commit()
        logger.info("Done!, you are now at: v{}", to_revert - 1)

    except Exception as error:
        logger.error(f"Failed to execute reversion script, due to {error}")
        logger.warning(
            "Changes are now rolled back, Please raise an issue regarding this. you are still in v{}",
            to_revert,
        )
        connection.rollback()
        log_migration(
            connection, to_revert, to_revert - 1, MigrationTrigger.CLI, False, error
        )
        connection.commit()
    finally:
        connection.close()


if __name__ == "__main__":
    migration(Path.cwd().parent.parent.parent / "TestResults" / "TeStReSuLtS.db")



from tortoise.models import Model
from tortoise.fields import (
    DatetimeField,
    IntField,
    FloatField,
    JSONField,
    CharEnumField,
    UUIDField,
    CharField,
    ReverseRelation,
    ForeignKeyField,
    BooleanField,
    ForeignKeyRelation,
    TextField,
)
from handshake.services.DBService.models.enums import (
    Status,
    SuiteType,
    RunStatus,
)


class CommandReportFields(Model):
    tests = IntField(
        default=0,
        null=False,
        description="No. of the test entries (either rolled up or direct)",
    )
    passed = IntField(
        default=0,
        null=False,
        description="Passed test entities (either rolled up or direct)",
    )
    failed = IntField(
        default=0,
        null=False,
        description="Failed test entities (either rolled up or direct)",
    )
    skipped = IntField(
        default=0,
        null=False,
        description="test entities that were skipped (either rolled up or direct)",
    )
    xfailed = IntField(
        default=0,
        null=False,
        description="test entities that were supposed to fail and has failed (either rolled up or direct)",
    )
    xpassed = IntField(
        default=0,
        null=False,
        description="test entities that were supposed to fail but has passed (either rolled up or direct)",
    )

    class Meta:
        abstract = True


class CommonDetailedFields(CommandReportFields):
    started = DatetimeField(descrption="Start Datetime field of test entity/session")
    ended = DatetimeField(
        null=True, description="End Datetime field of test entity/session"
    )
    duration = FloatField(
        default=0, null=False, description="duration of test entity/session"
    )

    class Meta:
        abstract = True


class EntityBaseSpecific:
    standing = CharEnumField(
        Status,
        description="observed status of the test run based on the test cases executed",
        default=Status.PENDING,
    )
    retried = IntField(
        default=0, null=False, description="number of the retries performed"
    )

    class Meta:
        abstract = True


class RunBase(CommonDetailedFields, EntityBaseSpecific):
    table = "RunBase"
    testID = UUIDField(pk=True)
    sessions = ReverseRelation["SessionBase"]
    tasks = ReverseRelation["TaskBase"]
    config = ReverseRelation["TestConfigBase"]
    started = DatetimeField(null=False, auto_now_add=True)
    projectName = CharField(
        max_length=30, null=False, description="Name of the project"
    )
    specStructure = JSONField(
        description="file structure of spec files", default=dict()
    )
    passedSuites = IntField(default=0, null=False, description="Passed suites")
    failedSuites = IntField(default=0, null=False, description="failed suites")
    skippedSuites = IntField(default=0, null=False, description="skipped suites")
    xpassedSuites = IntField(
        default=0, null=False, description="Was expected to failed but Passed suites"
    )
    xfailedSuites = IntField(
        default=0, null=False, description="Was expected to fail and was failed suites"
    )
    suites = IntField(default=0, null=False, description="total test suites")

    exitCode = IntField(
        null=False, default=0, description="Exit code for the test execution"
    )
    status = CharEnumField(
        RunStatus,
        description="status of the test run marked by the test framework",
        default=RunStatus.PENDING,
    )


class SessionBase(CommonDetailedFields):
    table = "SessionBase"
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="sessions", to_field="testID"
    )
    suites = ReverseRelation["SuiteBase"]
    sessionID = UUIDField(pk=True)
    entityName = CharField(max_length=30, default="")
    entityVersion = CharField(max_length=20, default="")
    simplified = TextField(
        default="", description="browser name & version &/ platform name included"
    )
    hooks = IntField(default=0, null=False, description="Number of hooks used")


class SuiteBase(EntityBaseSpecific, CommonDetailedFields):
    table = "SuiteBase"
    # https://tortoise.github.io/models.html#the-db-backing-field
    # so we require session_id instead of sessionID
    session: ForeignKeyRelation[SessionBase] = ForeignKeyField(
        "models.SessionBase", related_name="suites", to_field="sessionID"
    )
    started = DatetimeField(description="Start time of the test entity", null=True)
    attachments = ReverseRelation["AttachmentBase"]
    retries = ReverseRelation["RetriedBase"]
    rolled_up = ReverseRelation["RollupBase"]

    suiteID = UUIDField(pk=True)
    suiteType = CharEnumField(
        SuiteType,
        description="Specifies whether if it is a test suite or test case",
        null=False,
    )
    title = TextField(max_length=225)
    file = TextField(max_length=150, null=False, description="path to the spec file")
    description = TextField(
        null=True, default="", description="Summary if provided for the test entity"
    )
    parent = CharField(max_length=45, description="Parent Suite's ID", default="")
    tags = JSONField(
        description="comma separated list of tags used by the framework to filter the suites or spec files",
        default=[],
    )
    modified = DatetimeField(
        auto_now=True, description="Modified timestamp", null=False
    )
    errors = JSONField(description="Errors found", default=[], null=True)
    setup_duration = FloatField(
        default=0, null=False, description="duration of test entity's setup process"
    )
    teardown_duration = FloatField(
        default=0, null=False, description="duration of test entity's teardown process"
    )
    retried_later = BooleanField(
        default=False,
        required=False,
        description="was this test entity retried later",
    )


class RollupBase(CommandReportFields):
    suite: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="rollup", to_field="suiteID"
    )


class RetriedBase(Model):
    tests = JSONField(
        default=[],
        description="details of all suites which were retried for a particular test case",
    )
    suite: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="retries", to_field="suiteID"
    )

    length = IntField(
        description="Just the length of the tests cols", default=1, null=False
    )

    modified = DatetimeField(
        auto_now=True, description="Modified timestamp", null=False
    )



from loguru import logger
from handshake.services.DBService.models.dynamic_base import TaskBase, JobType
from handshake.services.DBService.models.result_base import RunBase, RunStatus
from uuid import uuid4, UUID
from typing import Union, List
from handshake.services.DBService.models.attachmentBase import (
    TestLogBase,
    LogType,
    LogGeneratedBy,
)


async def register_patch_suite(suiteID: str, testID: str, connection=None):
    _, created = await TaskBase.get_or_create(
        ticketID=suiteID, test_id=testID, type=JobType.MODIFY_SUITE, using_db=connection
    )
    if created:
        return _
    return False


async def register_patch_test_run(testID: str, connection=None) -> TaskBase:
    _, created = await TaskBase.get_or_create(
        type=JobType.MODIFY_TEST_RUN,
        test_id=testID,
        ticketID=testID,
        using_db=connection,
    )
    return _


async def mark_as_interrupted(test_id: str):
    record = await RunBase.filter(testID=test_id).first()
    record.status = RunStatus.INTERRUPTED
    await record.save()


async def register_bulk_patch_suites(
    testID: str,
    suites: List[str],
    connection=None,
) -> List[TaskBase]:
    tasks = await TaskBase.bulk_create(
        [
            TaskBase(ticketID=suite, test_id=testID, type=JobType.MODIFY_SUITE)
            for suite in suites
        ],
        100,
        using_db=connection,
    )
    return tasks


async def register_bulk_excel_export(
    testIDs: List[str],
    connection=None,
) -> List[TaskBase]:
    tasks = await TaskBase.bulk_create(
        [
            TaskBase(ticketID=str(uuid4()), test_id=test, type=JobType.EXPORT_EXCEL)
            for test in testIDs
        ],
        100,
        using_db=connection,
    )
    return tasks


async def _mark_custom_task(reason: str, job_type: JobType, test_id: str):
    logger.warning(reason)
    await TaskBase.create(ticketID=str(uuid4()), type=job_type, test_id=test_id)


async def mark_for_prune_task(test_id: str):
    # someone called this explicitly, hence it's a warning
    await _mark_custom_task(
        "Requested to prune some tasks", JobType.PRUNE_TASKS, test_id
    )


async def cancel_patch_for_test_run(
    test_id: Union[str, UUID],
    reason: str,
    was_interrupted: bool,
    generated_by: str,
    **extra,
) -> False:
    logger.error(reason)
    await TestLogBase.create(
        test_id=str(test_id),
        title="Cancelled patch for a test run",
        message=reason,
        type=LogType.ERROR,
        feed=extra,
        generatedByGroup=LogGeneratedBy.SCHEDULER,
        generatedBy=generated_by,
    )
    await mark_for_prune_task(test_id)
    if was_interrupted:
        await mark_as_interrupted(test_id)

    return False


async def warn_about_test_run(
    test_id: Union[str, UUID, List[UUID]],
    about: str,
    generated_by: str,
    bulk=False,
    **extra,
) -> True:
    logger.warning(about)
    if not bulk:
        await TestLogBase.create(
            test_id=str(test_id),
            title="A Warning was Raised for this test run",
            message=about,
            type=LogType.WARN,
            feed=extra,
            generatedByGroup=LogGeneratedBy.SCHEDULER,
            generatedBy=generated_by,
        )
    else:
        await TestLogBase.bulk_create(
            [
                TestLogBase(
                    test_id=str(_),
                    title="A Warning was Raised for this test run",
                    message=about,
                    type=LogType.WARN,
                    feed=extra,
                    generatedByGroup=LogGeneratedBy.SCHEDULER,
                    generatedBy=generated_by,
                )
                for _ in test_id
            ]
        )
    return True



from handshake.services.DBService.lifecycle import close_connection
from sanic.blueprints import Blueprint
from sanic.response import HTTPResponse, text
from sanic.request import Request

one_liners = Blueprint(name="one_liners", url_prefix="/")


# Here we will have requests which are very simple but used at crucial points


@one_liners.get("/")
async def health_status(request: Request) -> HTTPResponse:
    return text("1", status=200)


# bye is a core request, so make sure to handle it carefully
@one_liners.post("/bye")
async def bye(request: Request) -> HTTPResponse:
    await close_connection()
    request.app.m.terminate()

    return text("1", status=202)



from handshake.services.DBService.models.result_base import (
    SessionBase,
    SuiteBase,
    RunBase,
)
from handshake.services.DBService.models.types import (
    RegisterSuite,
    PunchInSuite,
    UpdateSuite,
    UpdateSession,
    PydanticModalForTestRunConfigBase,
    PydanticModalForTestRunUpdate,
    WrittenAttachmentForEntity,
    MarkTestRun,
)
from handshake.services.Endpoints.blueprints.utils import (
    prune_nones,
)
from handshake.services.Endpoints.define_api import definition
from handshake.services.DBService.models.config_base import TestConfigBase
from handshake.services.DBService.models.enums import (
    Status,
    SuiteType,
)
from handshake.services.DBService.models.static_base import StaticBase
from sanic.blueprints import Blueprint
from sanic.response import JSONResponse, text, HTTPResponse
from sanic.log import error_logger
from sanic.request import Request
from handshake.services.DBService.shared import get_test_id
from handshake.services.SchedularService.register import (
    register_patch_suite,
    register_patch_test_run,
)
from handshake.services.DBService.lifecycle import attachment_folder, db_path
from tortoise.expressions import F

update_service = Blueprint("UpdateService", url_prefix="/save")


@update_service.put("/PunchInSuite", error_format="json")
@definition(
    summary="Updates the suite's start datetime",
    description="we have dynamic dashboard, and since the start time can be provided anytime from the framework, "
    "we allow suite to be created without starttime, this request is to update start time as soon as "
    "possible.",
    tag="update",
    body={"application/json": PunchInSuite.model_json_schema()},
)
async def punch_in_test_suite(request: Request) -> HTTPResponse:
    payload = request.json
    suite = PunchInSuite.model_validate(payload)
    suite_record = await SuiteBase.filter(suiteID=suite.suiteID).first()
    if not suite_record:
        error_logger.error("Was not able to found {} suite", str(suite.suiteID))
        return text(f"Suite {suite.suiteID} was not found", status=404)

    payload = suite.model_dump()
    payload["standing"] = Status.PROCESSING
    suite_record = await suite_record.update_from_dict(payload)
    await suite_record.save()
    return text(str(suite_record.suiteID), status=200)


@update_service.put("/Suite", error_format="json")
@definition(
    summary="Updates the suite past the test suite's execution",
    description="Once the Test suite/test gets executed, through this endpoint "
    "we would updates its status and timings on the registered suite/test",
    tag="update",
    body={"application/json": UpdateSuite.model_json_schema()},
)
async def update_suite_details(request: Request) -> HTTPResponse:
    payload = prune_nones(request.json)

    suite = UpdateSuite.model_validate(payload)
    suite_record = await SuiteBase.filter(suiteID=suite.suiteID).first()
    if not suite_record:
        error_logger.error("Was not able to found {} suite", str(suite.suiteID))
        return text(f"Suite {suite.suiteID} was not found", status=404)

    # first, we save the details that were provided
    before_duration = suite_record.duration or 0
    if suite_record.suiteType == SuiteType.SUITE:
        suite.standing = Status.YET_TO_CALCULATE
    else:
        note = {
            suite.standing.lower(): 1,
            "tests": 1,
        }

        if (
            suite_record.suiteType == SuiteType.TEARDOWN
            or suite_record.suiteType == SuiteType.SETUP
        ):
            note[f"{suite_record.suiteType.lower()}_duration"] = suite.duration
        await suite_record.update_from_dict(note)

    await suite_record.update_from_dict(prune_nones(suite.model_dump()))
    await suite_record.save()

    # now we calculate certain data
    added_task = False

    match suite_record.suiteType:
        case SuiteType.SUITE:
            added_task = bool(
                await register_patch_suite(suite_record.suiteID, get_test_id())
            )
        # we can't have combined expression with duration and int at the same time, so we update twice
        case SuiteType.SETUP:
            # note here suite_record.duration means hook's duration
            await SuiteBase.filter(suiteID=suite_record.parent).update(
                setup_duration=F("setup_duration") + suite_record.duration
            )
            await SuiteBase.filter(suiteID=suite_record.parent).update(
                setup_duration=F("setup_duration") - before_duration
            )
        case SuiteType.TEARDOWN:
            await SuiteBase.filter(suiteID=suite_record.parent).update(
                teardown_duration=F("teardown_duration") + suite_record.duration
            )
            await SuiteBase.filter(suiteID=suite_record.parent).update(
                teardown_duration=F("teardown_duration") - before_duration
            )

    return text(str(suite_record.suiteID), status=201 if added_task else 200)


@update_service.put("/Session", error_format="json")
@definition(
    summary="Updates the session past the test session's execution",
    description="Once the Test session gets executed, through this endpoint "
    "we would updates its status and timings on the registered session",
    tag="update",
    body={"application/json": UpdateSession.model_json_schema()},
)
async def update_test_session_details(request: Request) -> HTTPResponse:
    session = UpdateSession.model_validate(request.json)
    test_session = await SessionBase.filter(sessionID=session.sessionID).first()
    if not test_session:
        error_logger.error("Expected {} session was not found", str(session.sessionID))
        return text(f"Session {session.sessionID} was not found", status=404)

    await test_session.update_from_dict(session.model_dump())
    await test_session.save()
    return text(f"{session.sessionID} was updated", status=200)


# NOTE: this API was made specifically to support a registering describeBlocks
# NOTE: Planning to depreciate this soon.
@update_service.put("/registerParentEntities")
@definition(
    summary="Registers a set of parent suites starting from root hierarchy",
    description="Registers set of parent suites with provided meta details of suites and session id",
    tag="register",
    body={"application/json": RegisterSuite},
)
async def register_parent_entities(request: Request) -> HTTPResponse:
    prev_parent = ""
    store = []
    for _suite in request.json:
        if isinstance(_suite, str):
            prev_parent = _suite
            store.append(prev_parent)
            continue

        suite = RegisterSuite.model_validate(_suite)
        values = suite.model_dump()
        values["parent"] = prev_parent
        suite_record = await SuiteBase.create(**values)
        prev_parent = str(suite_record.suiteID)
        await suite_record.save()
        store.append(prev_parent)

    return JSONResponse(body=store, status=201)


@update_service.put("/Run")
async def update_test_run(request: Request) -> HTTPResponse:
    to_update = MarkTestRun.model_validate(request.json)
    test_id = get_test_id()

    await register_patch_test_run(test_id)
    record = await RunBase.filter(testID=test_id).first()

    await record.update_from_dict(to_update.model_dump())
    await record.save()
    return text("updated test run successfully", status=200)


@update_service.put("/currentRun")
async def update_run_config(request: Request) -> HTTPResponse:
    run_config = PydanticModalForTestRunConfigBase.model_validate(request.json)
    config = await TestConfigBase.create(
        test_id=get_test_id(),
        platform=run_config.platformName,
        framework=run_config.framework,
        maxInstances=run_config.maxInstances,
        fileRetries=run_config.fileRetries,
        avoidParentSuitesInCount=run_config.avoidParentSuitesInCount,
        bail=run_config.bail,
        tags=run_config.tags,
    )

    test = await config.test
    await test.update_from_dict(dict(exitCode=run_config.exitCode))
    await test.save()

    return text("provided config was saved successfully.", status=200)


@update_service.put("/updateTestRun")
async def update_run(request: Request) -> HTTPResponse:
    about_run = PydanticModalForTestRunUpdate.model_validate(request.json)
    await register_patch_test_run(get_test_id())
    if not about_run:
        return text("No changes were made.", status=400)

    test = await RunBase.filter(testID=get_test_id()).first()
    await test.update_from_dict(about_run.model_dump())
    await test.save()
    return text(
        "test run was updated and task was also added successfully.", status=200
    )


@update_service.put("/registerAWrittenAttachment", error_format="json")
@definition(
    summary="Attach an Images to the specified entity",
    description="Writes an attachment inside of our Attachments folder, and attaches it with the specified entity",
    tag="add",
    body={"application/json": WrittenAttachmentForEntity},
)
async def saveImage(request: Request) -> HTTPResponse:
    attachment = WrittenAttachmentForEntity.model_validate(request.json)
    record = await StaticBase.create(**attachment.model_dump())
    file_name = f"{record.attachmentID}.{record.type.lower()}"
    record.value = file_name
    await record.save()
    # we can save the file in this request itself, but no. we let the framework's custom reporter cook.
    return text(
        str(
            attachment_folder(db_path())
            / get_test_id()
            / str(attachment.entity_id)
            / file_name
        ),
        status=201,
    )



from handshake.services.DBService.models.result_base import SuiteBase, RunBase
from tortoise.models import Model
from tortoise.fields import (
    ForeignKeyField,
    ForeignKeyRelation,
    TextField,
    IntField,
    DatetimeField,
    CharField,
    BooleanField,
    CharEnumField,
    JSONField,
    IntEnumField,
)
from handshake.services.DBService.models.enums import LogType, LogGeneratedBy


class AssertBase(Model):
    entity: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="assertion", to_field="suiteID"
    )
    passed = BooleanField(
        default=False, description="Whether the assertion passed or not"
    )
    wait = IntField(
        description="Number of milli-seconds configured to wait for this test",
        default=-1,
        null=True,
    )
    interval = IntField(
        description="interval (in milli-seconds) to test this assertion until it passes",
        default=-1,
        null=True,
    )
    title = TextField(description="Name of the Assertion")
    message = TextField(description="Message attached to the assertion")


class CommonLogBase(Model):
    type = CharEnumField(LogType, description="Log type", null=False)
    title = TextField(description="title for the log", null=False)
    message = TextField(description="formatted log message", default="", null=True)
    feed = JSONField(
        description="If you want to share any feed here", null=True, default={}
    )
    generatedBy = CharField(
        max_length=80,
        description="which job generated this, Null if generated by the user",
        null=False,
    )
    generatedByGroup = IntEnumField(
        LogGeneratedBy,
        default=LogGeneratedBy.USER,
        description="generated by api/scheduler jobs/user",
    )
    tags = JSONField(
        description="comma separated list of tags used by the framework to filter the suites or spec files",
        default=[],
        null=True,
    )
    dropped = DatetimeField(auto_now=True, description="timestamp", null=False)

    class Meta:
        abstract = True


class EntityLogBase(CommonLogBase):
    entity: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="entityLog", to_field="suiteID"
    )


class TestLogBase(CommonLogBase):
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="log", to_field="testID", pk=True
    )



import base64
from sanic.blueprints import Blueprint
from sanic.response import text, HTTPResponse
from sanic.log import logger
from sanic.request import Request
from handshake.services.DBService.models.types import (
    AddAttachmentForEntity,
)
from handshake.services.DBService.shared import root_dir
from handshake.services.DBService.models.static_base import (
    StaticBase,
)
from handshake.services.SchedularService.constants import writtenAttachmentFolderName
from handshake.services.Endpoints.define_api import definition

writeServices = Blueprint("WriteService", url_prefix="/write")


@writeServices.put("/addAttachmentForEntity", error_format="json")
@definition(
    summary="adds an image to the specified entity",
    description="saves an image and then attaches it to a specified entity",
    tag="add",
    deprecated=True,
    body={"application/json": AddAttachmentForEntity},
)
async def saveImage(request: Request) -> HTTPResponse:
    attachment = AddAttachmentForEntity.model_validate(request.json)

    record = await StaticBase.create(
        entity_id=attachment.entityID,
        description=attachment.description,
        type=attachment.type,
        title=attachment.title,
    )

    file_name = f"{record.attachmentID}.{record.type.lower()}"
    logger.debug("Received a file from user, saving it as {}", file_name)
    root = root_dir() / writtenAttachmentFolderName
    root.mkdir(exist_ok=True)
    test_root = root / str((await (await record.entity).session).test_id)
    test_root.mkdir(exist_ok=True)

    file = test_root / file_name
    file.write_bytes(base64.b64decode(attachment.value))
    record.value = file_name
    await record.save()
    return text(str(record.attachmentID), status=201)



from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.result_base import SuiteBase, Status, SuiteType
from handshake.services.SchedularService.modifySuites import patchTestSuite
from handshake.services.SchedularService.constants import JobType
from handshake.services.SchedularService.completeTestRun import patchTestRun
from handshake.services.SchedularService.flag_tasks import pruneTasks
from loguru import logger
from tortoise.expressions import Q, Subquery
from tortoise.functions import Min
from asyncio import TaskGroup
from pathlib import Path
from handshake.services.SchedularService.register import cancel_patch_for_test_run
from handshake.Exporters.excel_exporter import excel_export, ExcelExporter
from click import progressbar


async def safety_checks():
    prune_task = await TaskBase.filter(type=JobType.PRUNE_TASKS).first()
    if prune_task:
        await pruneTasks(prune_task.ticketID)


async def get_tasks():
    return (
        await TaskBase.filter(
            Q(picked=False)
            & Q(processed=False)
            & Q(processed=False)
            & Q(type=JobType.MODIFY_SUITE)
        )
        .order_by("dropped")  # old is on top
        .only("ticketID")
        .values_list("ticketID", flat=True)
    )


async def patch_jobs(include_excel_export: bool = False, db_path: Path = None):
    tasks = await get_tasks()
    first_time = True
    before = len(tasks)

    with progressbar(length=before, label="Processing Test Suites") as bar:
        while True:
            await safety_checks()

            async with TaskGroup() as patcher:
                # list of tasks which were not picked and processed and are specific to MODIFY_SUITE
                tasks = tasks if first_time else (await get_tasks())
                bar.update(before - (len(tasks) if tasks else 0))

                if first_time:
                    first_time = False

                if not tasks:
                    break

                # providers are list of child suites that needs to be processed before processing their parents
                providers = SuiteBase.filter(
                    Q(parent__in=tasks)
                    & (
                        Q(standing=Status.YET_TO_CALCULATE)
                        | Q(standing=Status.PROCESSING)
                    )
                    & Q(suiteType=SuiteType.SUITE)
                )
                # these are the list of the suites that needs to be processed and can be processed
                required_suites = SuiteBase.filter(
                    Q(suiteID__in=tasks)
                    & ~Q(suiteID__in=Subquery(providers.values("parent")))
                )

                # get a list of tasks which do not have children that are yet to process
                to_process = await TaskBase.filter(
                    Q(
                        ticketID__in=Subquery(
                            SuiteBase.filter(
                                # these are the list of suites that can be processed
                                # but filters out the suites whose previous retries were not processed
                                Q(
                                    suiteID__in=Subquery(
                                        required_suites.values("suiteID")
                                    )
                                )
                                & Q(
                                    retried=Subquery(
                                        required_suites.only("retried")
                                        .annotate(minRetries=Min("retried"))
                                        .first()
                                        .values("minRetries")
                                    )
                                )
                            ).values("suiteID")
                        )
                    )
                ).all()

                # so we process the suites whose child suites are all processed and their previous retries were
                # processed

                # if there are no such tasks to process, then check if the tasks are there for the child suites
                if not (
                    to_process
                    or await TaskBase.filter(
                        Q(ticketID=Subquery(providers.values("suiteID")))
                    ).exists()
                ):
                    # if so, mark each of the tests runs as banned.
                    for task in await TaskBase.filter(Q(ticketID__in=tasks)).all():
                        await cancel_patch_for_test_run(
                            task.test_id,
                            "Failed to find a way to process a parent suite, as the child suite was not registered.",
                            True,
                            "patcher-for-next-test-run-patch",
                            parent_suite=task.ticketID,
                            job=task.type,
                        )

                for task in to_process:
                    task.picked = True
                    await task.save()

                    patcher.create_task(
                        patchTestSuite(task.ticketID, task.test_id),
                        name=task.ticketID,
                    )

                # NOTE: make sure to pick the task before adding a new task

        bar.update(before if before else 1)

    logger.debug("Processing Test Runs")
    async with TaskGroup() as patcher:
        for job in await TaskBase.filter(
            Q(type=JobType.MODIFY_TEST_RUN) & Q(picked=False) & Q(processed=False)
        ).all():
            job.picked = True
            await job.save()
            patcher.create_task(patchTestRun(job.ticketID), name=job.ticketID)

    if not (db_path and include_excel_export and excel_export):
        logger.debug("Done!")
        return

    logger.debug("Exporting Test Runs")
    async with TaskGroup() as patcher:
        for job in await TaskBase.filter(
            Q(type=JobType.EXPORT_EXCEL)
            & Q(picked=False)
            & Q(processed=False)
            & ~Q(test__standing=Status.PENDING)
        ).all():
            exporter = ExcelExporter(db_path)
            job.picked = True
            await job.save()
            patcher.create_task(
                exporter.start_exporting(job.test_id), name=job.ticketID
            )

    logger.info("Done!")



from handshake.services.DBService.models.result_base import SessionBase, SuiteBase
from handshake.services.DBService.models.attachmentBase import (
    AssertBase,
    EntityLogBase,
    LogGeneratedBy,
)
from handshake.services.DBService.models.static_base import (
    AttachmentBase,
    AttachmentType,
)
from handshake.services.DBService.models.types import (
    AddAttachmentForEntity,
    RegisterSession,
    CreatePickedSuiteOrTest,
    PydanticModalForCreatingTestRunConfigBase,
    Status,
    SuiteType,
)
from pydantic import ValidationError
from handshake.services.Endpoints.define_api import definition
from handshake.services.DBService.models.config_base import TestConfigBase
from sanic.blueprints import Blueprint
from sanic.response import text, HTTPResponse
from sanic.log import error_logger
from sanic.request import Request
from handshake.services.DBService.shared import get_test_id
from handshake.services.SchedularService.register import register_bulk_patch_suites
from typing import List
from handshake.services.Endpoints.blueprints.utils import (
    attachWarn,
    extractPydanticErrors,
)

create_service = Blueprint("CreateService", url_prefix="/create")


@create_service.post("/Session")
@definition(
    summary="Registers a Session",
    description="Registers a session with datetime on the currently running Test Run. Please take a note of the "
    "sessionID sent",
    tag="create",
    body={"application/json": RegisterSession.model_json_schema()},
)
async def register_test_session(request: Request) -> HTTPResponse:
    try:
        session = RegisterSession.model_validate(request.json)
        session_record = await SessionBase.create(
            **session.model_dump(), test_id=get_test_id()
        )
        await session_record.save()
    except Exception as error:
        error_logger.error(
            "Failed to create a session due to exception: {}", str(error)
        )
        return text(str(error), status=400)

    return text(str(session_record.sessionID), status=201)


@create_service.post("/RunConfig")
async def create_test_run_config(request: Request) -> HTTPResponse:
    run_config = PydanticModalForCreatingTestRunConfigBase.model_validate(request.json)
    _, created = await TestConfigBase.get_or_create(
        test_id=get_test_id(),
        **run_config.model_dump(),
    )

    return text(
        (
            "configuration was added to the current test run"
            if created
            else "config cannot be updated, please raise a request for this endpoint"
        ),
        status=201 if created else 406,
    )


@create_service.post("/Suite")
async def create_suite(request: Request) -> HTTPResponse:
    suite = CreatePickedSuiteOrTest.model_validate(request.json)
    to_load = suite.model_dump()
    if to_load.pop("is_processing"):
        to_load["standing"] = Status.PROCESSING
    else:
        to_load["standing"] = Status.PENDING

    suite_record = await SuiteBase.create(**to_load)
    await suite_record.save()

    return text(str(suite_record.suiteID), status=201)


@create_service.post("/ScheduleSuites")
async def register_modify_suites(request: Request) -> HTTPResponse:
    test_id = get_test_id()

    suites_to_register = await SuiteBase.filter(
        session__test_id=test_id,
        suiteType=SuiteType.SUITE,
        started=None,
        ended=None,
    )

    suites = []
    for suite in suites_to_register:
        suite.standing = str(Status.YET_TO_CALCULATE)
        suites.append(suite.suiteID)

    suites_to_register and await SuiteBase.bulk_update(
        suites_to_register, ("standing",), 100
    )
    await register_bulk_patch_suites(test_id, suites)

    return text("Done", status=202)


@create_service.post("/Attachments")
@definition(
    summary="adds multiple attachments to the specified entities",
    description="provide the list of attachments (assertion/link/description) as mentioned in the body, "
    "and attachments would be inserted in their respective tables",
    tag="add",
    body={"application/json": List[AddAttachmentForEntity]},
)
async def addAttachmentForEntity(request: Request) -> HTTPResponse:
    attachments = []
    note = []
    logs = []
    assertions = []

    for _ in request.json:
        try:
            if _["type"] == AttachmentType.ASSERT:
                value = _.get("value", dict())
                value["wait"] = value.get("wait", -1)
                value["interval"] = value.get("interval", -1)

            attachment = AddAttachmentForEntity.model_validate(_)
        except ValidationError as error:
            note.append(_.get("entityID", False))
            await attachWarn(extractPydanticErrors(request.url, _, error), request.url)
            continue

        match attachment.type:
            case AttachmentType.ASSERT:
                assertions.append(
                    await AssertBase(
                        **dict(
                            entity_id=attachment.entity_id,
                            title=attachment.title,
                            message=attachment.description,
                            passed=attachment.value["passed"],
                            interval=attachment.value["interval"],
                            wait=attachment.value["wait"],
                        )
                    )
                )

            case AttachmentType.LOG:
                logs.append(
                    await EntityLogBase(
                        **dict(
                            entity_id=attachment.entity_id,
                            title=attachment.title,
                            message=attachment.description,
                            type=attachment.value["type"],
                            tags=attachment.tags,
                            feed=attachment.extraValues,
                            generatedByGroup=LogGeneratedBy.USER,
                            generatedBy="User",
                        )
                    )
                )

            case _:
                attachments.append(AttachmentBase(**attachment.model_dump()))

    if attachments:
        await AttachmentBase.bulk_create(attachments)
    if logs:
        await EntityLogBase.bulk_create(logs)
    if assertions:
        await AssertBase.bulk_create(assertions)
    return text(
        "Attachments was added successfully", status=201 if not len(note) else 206
    )



import datetime
from typing import Optional
from asyncio import gather
from handshake import __version__
from handshake.services.DBService.lifecycle import (
    handshake_meta,
    save_handshake_meta_dashboard,
    handshake_meta_dashboard,
    attachment_folder,
)
from handshake.services.SchedularService.constants import writtenAttachmentFolderName
from handshake.Exporters.json_exporter import JsonExporter
from shutil import rmtree, copytree
from loguru import logger
from pathlib import Path
from httpx import AsyncClient, RequestError, StreamError
from tarfile import open as t_open
from aiofiles import open as a_open
from click import progressbar


class HTMLExporter(JsonExporter):
    template: Path = Path(__file__).parent / "dashboard.tar.bz2"
    fetch: bool
    fetched: bool = False

    def __init__(self, db_path: Path, save_in: Path, dev_run: bool = False):
        super().__init__(db_path, save_in, dev_run)
        self.html_export_in = save_in

    def prepare(self):
        # we reset entire export folder
        if self.html_export_in.exists():
            logger.debug("removing previous results")
            rmtree(self.html_export_in)
        else:
            self.html_export_in.mkdir()

        self.fetch = not self.template.exists()
        super().prepare()

    async def start_exporting(
        self,
        run_id: Optional[str] = None,
        skip_project_summary: bool = False,
        skip_prepare: bool = False,
    ):
        self.prepare()
        await gather(
            self.prepare_download(),
            super().start_exporting(run_id, skip_project_summary, True),
        )

    async def prepare_download(self):
        meta = handshake_meta_dashboard()
        file_ready = False

        if not self.fetch:
            if not meta:
                self.fetch = True  # if the meta is not available, we fetch the file
            elif meta and meta.get("version", "") != __version__:
                self.fetch = True  # if the version is not matching, we fetch the file
            else:
                file_ready = True

        if self.fetch:
            file_ready = await self.download_zip()

        if file_ready:
            with t_open(self.template, "r:bz2") as tar_file:
                tar_file.extractall(self.html_export_in)

            logger.info(
                'HTML Report is available in {}, Please execute handshake display "{}" at {}',
                self.html_export_in,
                self.html_export_in.name,
                self.html_export_in.parent,
            )
        else:
            logger.warning(
                "HTML Export was faulted due to failure in downloading the zip file"
            )

    @staticmethod
    async def download_zip() -> bool:
        """
        Downloads a zip file from a given URL and saves it to the specified location.

        Returns:
        bool: True if the download is successful, False otherwise.
        """
        url = handshake_meta()["0"].get("browser_download_url", "")
        downloaded = False

        try:
            async with AsyncClient(follow_redirects=True) as client:
                async with client.stream("GET", url) as response:
                    with progressbar(
                        length=int(response.headers.get("Content-Length")),
                        label="Downloading Build",
                    ) as bar:
                        bar.length = 23
                        async with a_open(HTMLExporter.template, "wb") as file:
                            chunk_count = 1
                            async for chunk in response.aiter_bytes(chunk_size=1024):
                                bar.update(1024 * chunk_count)
                                chunk_count += 1
                                await file.write(chunk)

            downloaded = True

        except RequestError:
            logger.exception(
                "Failed to download from {}. due to the request error",
                url,
            )
        except StreamError:
            logger.exception(
                "Failed to download from {}. due to the stream error",
                url,
            )

        if downloaded:
            save_handshake_meta_dashboard(
                version=__version__,
                browser_download_url_for_dashboard=url,
                downloaded_dashboard_at=datetime.datetime.now().isoformat(),
            )

        return downloaded

    def completed(self):
        logger.info("Adding Attachments to the HTML Report.")
        copytree(
            attachment_folder(self.db_path),
            self.html_export_in / writtenAttachmentFolderName,
        )
        logger.info("Added Attachments to the HTML Report.")



import uuid
from typing import List, Optional, Union, Dict, Any
from handshake.services.DBService.models.enums import (
    Status,
    SuiteType,
    AttachmentType,
    RunStatus,
    LogType,
)
from pydantic import BaseModel
from datetime import datetime
from typing_extensions import TypedDict


# send an iso-format for the datetime string


class CommonRegisterCols(BaseModel):
    started: datetime


class Tag(TypedDict):
    label: str
    desc: str


class RegisterSession(CommonRegisterCols): ...


class RegisterSuite(CommonRegisterCols):
    title: str
    retried: int
    description: Optional[str] = ""
    suiteType: SuiteType
    expected: Optional[Status] = Status.PASSED
    session_id: uuid.UUID
    file: str
    parent: str
    tags: Optional[List[Tag]] = []


class CreatePickedSuiteOrTest(BaseModel):
    title: str
    retried: Optional[int] = 0
    started: Optional[datetime] = None
    description: Optional[str] = ""
    suiteType: SuiteType
    session_id: uuid.UUID
    file: str
    parent: str
    tags: Optional[List[Tag]] = []
    is_processing: Optional[bool] = True


class MarkSession(BaseModel):
    duration: float
    skipped: int
    passed: int
    failed: int
    tests: int
    hooks: int
    ended: datetime
    sessionID: uuid.UUID
    entityName: str
    entityVersion: str
    simplified: str


class UpdateSession(BaseModel):
    duration: float
    skipped: int
    passed: int
    failed: int
    tests: int
    hooks: int
    ended: datetime
    sessionID: uuid.UUID
    entityName: str
    entityVersion: str
    simplified: str


class Error(TypedDict):
    name: str
    message: str
    stack: Optional[str]


class UpdateSuite(BaseModel):
    suiteID: uuid.UUID
    duration: float
    started: Optional[datetime] = None
    expected: Optional[Status] = Status.PASSED
    ended: datetime
    errors: Optional[List[Error]] = []
    standing: Status


class PunchInSuite(CommonRegisterCols):
    suiteID: uuid.UUID


class MarkSuite(BaseModel):
    duration: float  # milliseconds
    ended: datetime
    suiteID: uuid.UUID
    errors: Optional[List[Error]] = []
    standing: Optional[Status] = Status.SKIPPED


class AssertionPayload(TypedDict):
    passed: bool
    wait: Optional[int]
    interval: Optional[int]


class LogPayload(TypedDict):
    type: LogType


# this is for all kinds of attachments
class AddAttachmentForEntity(BaseModel):
    entity_id: uuid.UUID
    type: AttachmentType
    description: Optional[str] = ""
    value: Union[AssertionPayload, LogPayload, str]
    title: Optional[str] = ""
    extraValues: Optional[Dict[Any, Any]] = {}
    tags: Optional[List[Tag]] = []


class WrittenAttachmentForEntity(BaseModel):
    entity_id: uuid.UUID
    type: AttachmentType
    description: Optional[str] = ""
    title: Optional[str] = ""


class ValueForTestRunConfigBase(TypedDict):
    version: str
    platformName: str


class PydanticModalForTestRunConfigBase(BaseModel):
    maxInstances: Optional[int] = 1
    platformName: str
    framework: str
    exitCode: int
    bail: Optional[int] = 0
    fileRetries: Optional[int] = 0
    avoidParentSuitesInCount: Optional[bool] = False
    tags: List[Tag]


class PydanticModalForCreatingTestRunConfigBase(BaseModel):
    maxInstances: Optional[int] = 1
    platform: str
    framework: str
    bail: Optional[int] = -1
    fileRetries: Optional[int] = 0
    avoidParentSuitesInCount: Optional[bool] = False
    tags: Optional[List[Tag]] = []


class PydanticModalForTestRunUpdate(CommonRegisterCols): ...


class MarkTestRun(BaseModel):
    exitCode: int
    status: RunStatus



from typing import TypedDict
from datetime import datetime
from handshake.services.DBService.models.enums import (
    Status,
    RunStatus,
)


class SummaryOfTestRun(TypedDict):
    testID: str
    passed: int
    failed: int
    skipped: int
    tests: int
    passedSuites: int
    failedSuites: int
    skippedSuites: int
    suites: int
    duration: int
    projectName: str
    exitCode: int
    status: RunStatus
    started: datetime
    ended: datetime
    retried: int
    standing: Status
